#include "eusart_lib.h"

// vars for UART ring buffers
#define USART_TX_BUFFER_SIZE			16											// size of UART RX ring buffer
#define USART_RX_BUFFER_SIZE			16											// size of UART RX ring buffer

/* size of RX/TX buffers */
#define USART_RX_BUFFER_MASK ( USART_RX_BUFFER_SIZE - 1)
#define USART_TX_BUFFER_MASK ( USART_TX_BUFFER_SIZE - 1)

#if ( USART_RX_BUFFER_SIZE & USART_RX_BUFFER_MASK )
#error RX buffer size is not a power of 2
#endif
#if ( USART_TX_BUFFER_SIZE & USART_TX_BUFFER_MASK )
#error TX buffer size is not a power of 2
#endif

// global variables
uint8_t   USART_TxBuffer[USART_TX_BUFFER_SIZE];										// UART-0 RX ring buffer
uint8_t   USART_TxBufferWr = 0;
uint8_t   USART_TxBufferRd = 0;
uint8_t   USART_TxBufferOverFlow = 0;
volatile uint8_t eusartTxBufferRemaining;

uint8_t   USART_RxBuffer[USART_RX_BUFFER_SIZE];										// UART-0 RX ring buffer
uint8_t   USART_RxBufferWr = 0;
uint8_t   USART_RxBufferRd = 0;
uint8_t   USART_RxBufferOverFlow = 0;
volatile uint8_t eusartRxCount;

void eusart_initialize(void) {
    // disable interrupts before changing states
    PIE1bits.RCIE = 0;
    PIE1bits.TXIE = 0;

    // Set the EUSART module to the options selected in the user interface.

    // ABDOVF no_overflow; SCKP async_noninverted_sync_fallingedge; RCIDL idle; BRG16 16bit_generator; WUE disabled; ABDEN disabled; 
    BAUDCON = 0x48;

    // SPEN enabled; OERR no_error; RX9 8-bit; RX9D 0x0; CREN enabled; ADDEN disabled; SREN disabled; FERR no_error; 
    RCSTA = 0x90;

    // TRMT TSR_empty; TX9 8-bit; TX9D 0x0; SENDB sync_break_complete; TXEN enabled; SYNC asynchronous; BRGH hi_speed; CSRC slave_mode; 
    TXSTA = 0x26;

    // Baud Rate = 9600; SPBRGL 207; 
    SPBRGL = 0xCF;

    // Baud Rate = 9600; SPBRGH 0; 
    SPBRGH = 0x00;

    // enable receive interrupt
    PIE1bits.RCIE = 1;
}

uint8_t eusart_getc(void) {
   uint8_t inval;

   while (USART_RxBufferRd == USART_RxBufferWr);						// wait until there are uint8_ts received

   PIE1bits.RCIE = 0;  //turn off interrupt while receiving
   inval = USART_RxBuffer[USART_RxBufferRd++];									// get byte
   if (USART_RxBufferRd >= sizeof(USART_RxBuffer)) USART_RxBufferRd = 0;	// wrap around
   eusartRxCount--;
   PIE1bits.RCIE = 1;  //turn back on interrupt

   return inval;
}

void eusart_putc(uint8_t txData) {
   uint8_t tmphead;
   uint8_t tmptail;

   if (0 == PIE1bits.TXIE) {
     TXREG = txData;
   } else {
     tmphead  = (USART_TxBufferWr + 1) & USART_TX_BUFFER_MASK;

      while ( tmphead == USART_TxBufferRd ){
         ;/* wait for free space in buffer */
      }
      PIE1bits.TXIE = 0;
      USART_TxBuffer[tmphead] = txData;
      USART_TxBufferWr = tmphead;
      eusartTxBufferRemaining--;
   }
   PIE1bits.TXIE = 1;
}

void eusart_gets( char *s )
{
  uint8_t ch = 0;

  ch = eusart_getc();

  while ( (ch != 0x0d ) && (ch != 0x0a ))
  {
    *s = ch;
    s++;
    ch = eusart_getc();
  }
  *s = 0x00;
}

void eusart_puts(const char *s )
{
    while (*s)
      eusart_putc(*s++);

}

void EUSART_Transmit_ISR(void) {
   uint8_t tmptail;

   if ( USART_TxBufferWr != USART_TxBufferRd)  //txbuffer not empty
   {
      // calculate and store new buffer index 
      tmptail = (USART_TxBufferRd + 1) & USART_TX_BUFFER_MASK;
      USART_TxBufferRd = tmptail;
      // get one byte from buffer and write it to UART 
      TXREG = USART_TxBuffer[tmptail];
        eusartTxBufferRemaining++;
   }
   else {
      PIE1bits.TXIE = 0;
   }
}

void EUSART_Receive_ISR(void) {
    if (1 == RCSTAbits.OERR) {
        // EUSART error - restart

        RCSTAbits.CREN = 0;
        RCSTAbits.CREN = 1;
    }

   // buffer overruns are flagged
   if( USART_RxBufferWr + 1 == USART_RxBufferRd ) // check overflow
      USART_RxBufferOverFlow = 1;	          // set overflow flag
   else
   {
      USART_RxBuffer[USART_RxBufferWr++] = RCREG;     // save new byte into the buffer
      if (USART_RxBufferWr >= sizeof(USART_RxBuffer)) USART_RxBufferWr = 0;   // wrap around
      eusartRxCount++;
   }
}

